UNIT abstract_syntax;

  INTERFACE

    USES standard;

    CONST id_size        = 9;

    TYPE  identifier     = STRING [id_size];

          a_process      = ^process_def;

          a_variable     = ^var_def;

          a_statement    = ^statement_def;

          an_expression  = ^expression_def;

          process_list   = ^proc_rec;

          proc_rec       = RECORD
                             head: a_process;
                             tail: process_list
                           END;

          var_list       = ^var_rec;

          var_rec        = RECORD
                             head: a_variable;
                             tail: var_list
                           END;

          stmt_list      = ^stmt_rec;

          stmt_rec       = RECORD
                             head: a_statement;
                             tail: stmt_list
                           END;

          process_def    = RECORD
                             name        : identifier;
                             number      : natural;
                             inputs      : process_list;
                             outputs     : process_list;
                             variables   : var_list;
                             process_vars: var_list;
                             body        : stmt_list;
                             unseen      : Boolean
                           END;

          var_def        = RECORD
                             name  : identifier;
                             offset: natural
                           END;

          stmt_kind      = (empty_stmt, assignment_stmt, while_stmt,    if_stmt,
                            send_stmt,  receive_stmt,    compound_stmt, gen_stmt
                           );

          an_assignment  = RECORD
                             variable: a_variable;
                             value   : an_expression
                           END;

          a_while        = RECORD
                             condition: an_expression;
                             statement: a_statement
                           END;

          an_if          = RECORD
                             condition: an_expression;
                             then_part: a_statement;
                             else_part: a_statement
                           END;

          proc_ref_kind  = (named_ref, proc_deref, any_ref);

          a_proc_ref     = RECORD
                            CASE kind: proc_ref_kind OF
                              named_ref : (the_ref: a_process );
                              proc_deref: (the_var: a_variable);
                              any_ref   : (the_any: a_variable)
                           END;

          a_send         = RECORD
                             value      : an_expression;
                             destination: a_proc_ref
                           END;

          a_receive      = RECORD
                             variable: a_variable;
                             source  : a_proc_ref
                           END;

          a_compound     = stmt_list;

          statement_def  = RECORD
                             unseen: Boolean;
                             CASE kind: stmt_kind OF
                               gen_stmt       : ((* "undefined" <statement> *));
                               empty_stmt     : (                             );
                               assignment_stmt: (the_assignment: an_assignment);
                               while_stmt     : (the_while     : a_while      );
                               if_stmt        : (the_if        : an_if        );
                               send_stmt      : (the_send      : a_send       );
                               receive_stmt   : (the_receive   : a_receive    );
                               compound_stmt  : (the_compound  : a_compound   )
                           END;

          a_binary_op    = (add_op, sub_op, or_op, mult_op, div_op, mod_op, and_op,
                            eq_op,  neq_op, ls_op, leq_op,  gtr_op, geq_op
                           );

          a_binary_expr  = RECORD
                             operator : a_binary_op;
                             operand_1: an_expression;
                             operand_2: an_expression
                           END;

          a_unary_op     = (plus_op, minus_op, not_op);

          a_unary_expr   = RECORD
                             operator: a_unary_op;
                             operand : an_expression
                           END;

          expr_kind      = (const_expr,  Bool_const_expr, var_expr,
                            binary_expr, unary_expr,      io_expr,  gen_expr
                           );

          expression_def = RECORD
                             unseen: Boolean;
                             CASE kind: expr_kind OF
                               gen_expr       : ((*  "undefined" <expr>     *));
                               const_expr     : (the_constant  : natural      );
                               Bool_const_expr: (the_Bool_const: Boolean      );
                               var_expr       : (the_variable  : a_variable   );
                               binary_expr    : (the_binary    : a_binary_expr);
                               unary_expr     : (the_unary     : a_unary_expr );
                               io_expr        : (the_io        : a_statement  )
                           END;

          a_program      = process_list;

    FUNCTION  new_process(name: identifier; inputs, outputs: process_list; variables,
                          process_vars: var_list; body: stmt_list
                         ): a_process;

    PROCEDURE add_process(p: a_process; VAR l: process_list);

    FUNCTION  process_ref(name: identifier; l: process_list): a_process;

    FUNCTION  new_variable(name: identifier): a_variable;

    PROCEDURE add_variable(v: a_variable; VAR l: var_list);

    FUNCTION  var_ref(name: identifier; l: var_list): a_variable;

    FUNCTION  new_empty_stmt: a_statement;

    FUNCTION  new_gen_stmt: a_statement;

    FUNCTION  new_assignment(variable: a_variable; value: an_expression): a_statement;

    FUNCTION  new_while(condition: an_expression; statement: a_statement): a_statement;

    FUNCTION  new_if(condition: an_expression; then_part, else_part: a_statement): a_statement;

    FUNCTION  new_send(value: an_expression; destination: a_proc_ref): a_statement;

    FUNCTION  new_receive(variable: a_variable; source: a_proc_ref): a_statement;

    FUNCTION  new_compound(l: stmt_list): a_statement;

    PROCEDURE add_statement(s: a_statement; VAR l: stmt_list);

    FUNCTION  new_gen_expr: an_expression;

    FUNCTION  new_const_expr(value: natural): an_expression;

    FUNCTION  new_var_expr(v: a_variable): an_expression;

    FUNCTION  new_Bool_const_expr(value: Boolean): an_expression;

    FUNCTION  new_binary_expr(operator: a_binary_op; operand_1, operand_2: an_expression): an_expression;

    FUNCTION  new_unary_expr(operator: a_unary_op; operand: an_expression): an_expression;

    FUNCTION  new_io_expr(operand: a_statement): an_expression;

  IMPLEMENTATION

    (*$R-*)

    FUNCTION  new_process(* name: identifier; inputs, outputs: process_list;
                            variables, process_vars: var_list; body: stmt_list
                         ): a_process *);
        VAR p: a_process;
    BEGIN
      new(p);
      p^.name        :=name;
      p^.inputs      :=inputs;
      p^.outputs     :=outputs;
      p^.variables   :=variables;
      p^.process_vars:=process_vars;
      p^.body        :=body;
      p^.unseen      :=false;
      new_process    :=p
    END;

    PROCEDURE add_process(* p: a_process; VAR l: process_list *);
    BEGIN
      IF l = nil THEN
      BEGIN
        new(l);
        l^.head:=p;
        l^.tail:=nil
      END
      ELSE add_process(p, l^.tail)
    END;

    FUNCTION process_ref(* name: identifier; l: process_list): a_process *);
    BEGIN
      IF l = nil THEN process_ref:=nil
      ELSE
      IF l^.head^.name = name THEN process_ref:=l^.head
      ELSE process_ref:=process_ref(name, l^.tail)
    END;

    FUNCTION  new_variable(* name: identifier): a_variable *);
        VAR v: a_variable;
    BEGIN
      new(v);
      v^.name:=name;
      new_variable:=v
    END;

    PROCEDURE add_variable(* v: a_variable; VAR l: var_list *);
    BEGIN
      IF l = nil THEN
      BEGIN
        new(l);
        l^.head:=v;
        l^.tail:=nil
      END
      ELSE add_variable(v, l^.tail)
    END;

    FUNCTION var_ref(* name: identifier; l: var_list): a_variable *);
    BEGIN
      IF l = nil THEN var_ref:=nil
      ELSE
      IF l^.head^.name = name THEN var_ref:=l^.head
      ELSE var_ref:=var_ref(name, l^.tail)
    END;

    FUNCTION  new_gen_stmt(*: a_statement *);
        VAR s: a_statement;
    BEGIN
      new(s);
      s^.unseen:=false;
      s^.kind:=gen_stmt;
      new_gen_stmt:=s
    END;

    FUNCTION  new_empty_stmt(*: a_statement *);
        VAR s: a_statement;
    BEGIN
      new(s);
      s^.unseen:=false;
      s^.kind:=empty_stmt;
      new_empty_stmt:=s
    END;

    FUNCTION  new_assignment(* variable: a_variable; value: an_expression): a_statement *);
        VAR s: a_statement;
    BEGIN
      new(s);
      s^.unseen:=false;
      s^.kind:=assignment_stmt;
      s^.the_assignment.variable:=variable;
      s^.the_assignment.value   :=value;
      new_assignment:=s
    END;

    FUNCTION  new_while(* condition: an_expression; statement: a_statement): a_statement *);
        VAR s: a_statement;
    BEGIN
      new(s);
      s^.unseen:=false;
      s^.kind:=while_stmt;
      s^.the_while.condition:=condition;
      s^.the_while.statement:=statement;
      new_while:=s
    END;

    FUNCTION  new_if(* condition: an_expression; then_part, else_part: a_statement): a_statement *);
        VAR s: a_statement;
    BEGIN
      new(s);
      s^.unseen:=false;
      s^.kind:=if_stmt;
      s^.the_if.condition:=condition;
      s^.the_if.then_part:=then_part;
      s^.the_if.else_part:=else_part;
      new_if:=s
    END;

    FUNCTION  new_send(* value: an_expression; destination: a_proc_ref): a_statement *);
        VAR s: a_statement;
    BEGIN
      new(s);
      s^.unseen:=false;
      s^.kind:=send_stmt;
      s^.the_send.value      :=value;
      s^.the_send.destination:=destination;
      new_send:=s
    END;

    FUNCTION  new_receive(* variable: a_variable; source: a_proc_ref): a_statement *);
        VAR s: a_statement;
    BEGIN
      new(s);
      s^.unseen:=false;
      s^.kind:=receive_stmt;
      s^.the_receive.variable:=variable;
      s^.the_receive.source  :=source;
      new_receive:=s
    END;

    FUNCTION  new_compound(* l: stmt_list): a_statement *);
        VAR s: a_statement;
    BEGIN
      new(s);
      s^.unseen:=false;
      s^.kind:=compound_stmt;
      s^.the_compound:=l;
      new_compound:=s
    END;

    PROCEDURE add_statement(* s: a_statement; VAR l: stmt_list *);
    BEGIN
      IF l = nil THEN
      BEGIN
        new(l);
        l^.head:=s;
        l^.tail:=nil
      END
      ELSE add_statement(s, l^.tail)
    END;

    FUNCTION  new_gen_expr(*: an_expression *);
        VAR e: an_expression;
    BEGIN
      new(e);
      e^.unseen:=false;
      e^.kind:=gen_expr;
      new_gen_expr:=e
    END;

    FUNCTION  new_const_expr(* value: natural): an_expression *);
        VAR e: an_expression;
    BEGIN
      new(e);
      e^.unseen:=false;
      e^.kind:=const_expr;
      e^.the_constant:=value;
      new_const_expr:=e
    END;

    FUNCTION  new_var_expr(* v: a_variable): an_expression *);
        VAR e: an_expression;
    BEGIN
      new(e);
      e^.unseen:=false;
      e^.kind:=var_expr;
      e^.the_variable:=v;
      new_var_expr:=e
    END;

    FUNCTION  new_Bool_const_expr(* value: Boolean): an_expression *);
        VAR e: an_expression;
    BEGIN
      new(e);
      e^.unseen:=false;
      e^.kind:=Bool_const_expr;
      e^.the_Bool_const:=value;
      new_Bool_const_expr:=e
    END;

    FUNCTION  new_binary_expr(* operator: a_binary_op; operand_1, operand_2: an_expression): an_expression *);
        VAR e: an_expression;
    BEGIN
      new(e);
      e^.unseen:=false;
      e^.kind:=binary_expr;
      e^.the_binary.operator :=operator;
      e^.the_binary.operand_1:=operand_1;
      e^.the_binary.operand_2:=operand_2;
      new_binary_expr:=e
    END;

    FUNCTION  new_unary_expr(* operator: a_unary_op; operand: an_expression): an_expression *);
        VAR e: an_expression;
    BEGIN
      new(e);
      e^.unseen:=false;
      e^.kind:=unary_expr;
      e^.the_unary.operator:=operator;
      e^.the_unary.operand :=operand;
      new_unary_expr:=e
    END;

    FUNCTION  new_io_expr(* operand: a_statement): an_expression *);
        VAR e: an_expression;
    BEGIN
      new(e);
      e^.unseen:=false;
      e^.kind:=io_expr;
      e^.the_io:=operand;
      new_io_expr:=e
    END;

END.
